//
// Copyright (c) 2019 INRIA
//

#ifndef __pinocchio_serialization_spatial_hpp__
#define __pinocchio_serialization_spatial_hpp__

#include "pinocchio/serialization/se3.hpp"
#include "pinocchio/serialization/motion.hpp"
#include "pinocchio/serialization/force.hpp"
#include "pinocchio/serialization/symmetric3.hpp"
#include "pinocchio/serialization/inertia.hpp"

#endif // ifndef __pinocchio_serialization_spatial_hpp__
